package com.itko.lisa.ext.trilead;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.itko.lisa.test.TestEvent;
import com.itko.lisa.test.TestExec;
import com.itko.lisa.test.TestRunException;
import com.trilead.ssh2.Connection;
import com.trilead.ssh2.SCPClient;

public class SCPEngine {
	
	private static Log log = LogFactory.getLog(SFTPEngine.class);

	/**
	 * 
	 * @param settings
	 * @param ts
	 * @return
	 * @throws Exception
	 */
	public String execute(SFTPStepSettings settings, TestExec ts) throws Exception {
		
		long lExecuteStart = System.currentTimeMillis();
		long lConnectionDone = 0;
		long lClientDone = 0;
		long lCommandDone = 0;
		long lCleanupDone = 0;
		
		// Parse the connection settings
		String tsHostname = ts.parseInState(settings.getHostname());
		String tsUsername = ts.parseInState(settings.getUsername());
		String tsPassword = ts.parseInState(settings.getPassword());
		String tsAuth = ts.parseInState(settings.getAuth());
		String tsPrivateKey = ts.parseInState(settings.getPrivateKey());
		String tsPrivateKeyPass = ts.parseInState(settings.getPrivateKeyPass());

		// Create the SFTP connection
		Connection conn = null;
		SCPClient client = null; 
		boolean invalidate = true;
		try {
			
			// Let's check to see if someone else has mad a Connection
			// to our same host that we can borrow, so we call getConnection()
			conn = TrileadConnectionFactory.getConnection(tsHostname, tsUsername, tsPassword, tsAuth, tsPrivateKey, tsPrivateKeyPass);
			lConnectionDone = System.currentTimeMillis();
			
			client = new SCPClient(conn);
			lClientDone = System.currentTimeMillis();
			
			// Parse the SFTPSettings
			String path = ts.parseInState(settings.getPath());
			String filename = ts.parseInState(settings.getFilename());
			String content = ts.parseInStateControlNesting(settings.getContent(), false);
			boolean saveToFile = settings.isSaveToFile();
			String localFilename = ts.parseInState(settings.getDestFile());

			// Some logging
			String msg = "Attempting to "
				+ settings.getCommand()
				+ " path=(" + path + ")"
				+ " file=(" + filename + ")";
			log.debug(msg);
			if (ts.getCurrentNode() != null) {
				ts.raiseEvent(TestEvent.EVENT_NODEMSG, ts.getCurrentNode().getName(), msg);
			}
			
			// Figure out the correct behavior
			String command = settings.getCommand();
			if (command.equalsIgnoreCase(SFTPStep.CMD_GET)) {
				msg = cmd_get(client, path, filename, saveToFile, localFilename);
			} else if (command.equalsIgnoreCase(SFTPStep.CMD_PUT)) {
				msg = cmd_put(client, path, filename, content);
			} else {
				log.error("Unknown SCP command: " + command);
				throw new RuntimeException("Unknown SCP command: " + command);
			}
			
			// Calculate the time
			lCommandDone = System.currentTimeMillis();
			
			// Cleanup and return
			invalidate = false;
			return msg;
		} catch (Exception e) {
			log.error(e.getMessage(), e);
			throw new TestRunException(e.getMessage(), e);
		} finally {
			TrileadConnectionFactory.cleanupConnection(conn, tsHostname, tsUsername, tsAuth, invalidate);
			
			// Make some calculations
			lCleanupDone = System.currentTimeMillis();
			long lConnection = lConnectionDone - lExecuteStart;
			long lClient = lClientDone - lConnectionDone;
			long lCommand = lCommandDone - lClientDone;
			long lCleanup = lCleanupDone - lCommandDone;
			
			if (ts.getCurrentNode() != null) {
				ts.raiseEvent(TestEvent.EVENT_LOGMSG, "SSH Connection Time", lConnection + "");
				ts.raiseEvent(TestEvent.EVENT_LOGMSG, "SFTP Client Creation Time", lClient + "");
				ts.raiseEvent(TestEvent.EVENT_EXECTIME, ts.getCurrentNode().getName(), lCommand + "");
				ts.raiseEvent(TestEvent.EVENT_LOGMSG, "SFTP Client Cleanup Time", lCleanup + "");
			}
		}
	}
	
//	/**
//	 * 
//	 * @param client
//	 * @param path
//	 * @param filename
//	 * @return
//	 * @throws IOException
//	 */
//	private String cmd_get(SCPClient client, String path, String filename, boolean saveToFile, String localFilename) throws IOException {
//		if (saveToFile) {
//			FileOutputStream out = new FileOutputStream(localFilename);
//			client.get(path + "/" + filename, out);
//			return "Local file: " + localFilename + " written.";
//		} else {
//			ByteArrayOutputStream fileOut = new ByteArrayOutputStream(); 
//			client.get(path + "/" + filename, fileOut);
//			return fileOut.toString();
//		}
//	}
	
	/**
    *
    *
    *
    * @param client
    *
    * @param path
    *
    * @param filenames
    *
    * @return
    *
    * @throws IOException
    */

   private String cmd_get(SCPClient client,
                          String path,
                          String filenames,
                          boolean saveToDir,
                          String localDir) throws IOException
   {

       String[] fileNameArray = filenames.split(",");
       String[] qualifiedFileNames = new String[fileNameArray.length];

       int i = 0;
       for ( String fileName : fileNameArray )
       {
           qualifiedFileNames[i] = path + "/" + fileName;
           i++;
       }

       File dir = new File(localDir);
       if ( !dir.exists() && !dir.mkdirs() )
       {
           throw new IOException("Cannot create dir's or dir with the given path : "
                   + path);
       }

       if ( saveToDir )
       {
           client.get(qualifiedFileNames, localDir);
           return "Local directory: " + localDir + " written.";

       }
       else
       {

           ByteArrayOutputStream fileOut = new ByteArrayOutputStream();

           client.get(path + "/" + filenames, fileOut);

           return fileOut.toString();

       }

   }
	
	/**
	 * 
	 * @param client
	 * @param path
	 * @param filename
	 * @param content
	 * @return
	 * @throws IOException
	 */
	private String cmd_put(SCPClient client, String path, String filename, String content) throws IOException {
		client.put(content.getBytes(), filename, path);
		return path + "/" + filename + " put.";
	}
}
