package com.itko.lisa.ext.trilead;

import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Date;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.itko.lisa.test.TestEvent;
import com.itko.lisa.test.TestExec;
import com.itko.lisa.test.TestRunException;
import com.trilead.ssh2.Connection;
import com.trilead.ssh2.SFTPv3Client;
import com.trilead.ssh2.SFTPv3DirectoryEntry;
import com.trilead.ssh2.SFTPv3FileHandle;

public class SFTPEngine {
	
	private static Log log = LogFactory.getLog(SFTPEngine.class);

	/**
	 * 
	 * @param settings
	 * @param ts
	 * @return
	 * @throws Exception
	 */
	public String execute(SFTPStepSettings settings, TestExec ts) throws Exception {
		
		long lExecuteStart = System.currentTimeMillis();
		long lConnectionDone = 0;
		long lClientDone = 0;
		long lCommandDone = 0;
		long lCleanupDone = 0;
		
		// Parse the connection settings
		String tsHostname = ts.parseInState(settings.getHostname());
		String tsUsername = ts.parseInState(settings.getUsername());
		String tsPassword = ts.parseInState(settings.getPassword());
		String tsAuth = ts.parseInState(settings.getAuth());
		String tsPrivateKey = ts.parseInState(settings.getPrivateKey());
		String tsPrivateKeyPass = ts.parseInState(settings.getPrivateKeyPass());

		// Create the SFTP connection
		Connection conn = null;
		SFTPv3Client client = null; 
		boolean invalidate = true;
		try {
			
			// Let's check to see if someone else has mad a Connection
			// to our same host that we can borrow, so we call getConnection()
			conn = TrileadConnectionFactory.getConnection(tsHostname, tsUsername, tsPassword, tsAuth, tsPrivateKey, tsPrivateKeyPass);
			lConnectionDone = System.currentTimeMillis();
			
			client = new SFTPv3Client(conn);
			lClientDone = System.currentTimeMillis();
			
			// Parse the SFTPSettings
			String path = ts.parseInState(settings.getPath());
			String filename = ts.parseInState(settings.getFilename());
			String content = ts.parseInStateControlNesting(settings.getContent(), false);
			boolean saveToFile = settings.isSaveToFile();
			String localFilename = ts.parseInState(settings.getDestFile());

			// Some logging
			String msg = "Attempting to "
				+ settings.getCommand()
				+ " path=(" + path + ")"
				+ " file=(" + filename + ")";
			log.debug(msg);
			if (ts.getCurrentNode() != null) {
				ts.raiseEvent(TestEvent.EVENT_NODEMSG, ts.getCurrentNode().getName(), msg);
			}
			
			// Figure out the correct behavior
			String command = settings.getCommand();
			if (command.equalsIgnoreCase(SFTPStep.CMD_LS)) {
				msg = cmd_ls(client, path);
			} else if (command.equalsIgnoreCase(SFTPStep.CMD_READ)) {
				msg = cmd_read(client, path, filename, saveToFile, localFilename);
			} else if (command.equalsIgnoreCase(SFTPStep.CMD_OVERWRITE)) {
				msg = cmd_overwrite(client, path, filename, content);
			} else if (command.equalsIgnoreCase(SFTPStep.CMD_FORCEWRITE)) {
				msg = cmd_forcewrite(client, path, filename, content);
			} else if (command.equalsIgnoreCase(SFTPStep.CMD_CREATE)) {
				msg = cmd_create(client, path, filename, content);
			} else if (command.equalsIgnoreCase(SFTPStep.CMD_DELETE)) {
				msg = cmd_delete(client, path, filename);
			} else {
				log.error("Unknown SFTP command: " + command);
				throw new RuntimeException("Unknown SFTP command: " + command);
			}
			
			// Calculate the time
			lCommandDone = System.currentTimeMillis();
			
			// Cleanup and return
			invalidate = false;
			return msg;
		} catch (Exception e) {
			log.error(e.getMessage(), e);
			throw new TestRunException(e.getMessage(), e);
		} finally {
			
			// Close the client
			if (client != null) {
				client.close();
			}
			
			// Toss the connection
			TrileadConnectionFactory.cleanupConnection(conn, tsHostname, tsUsername, tsAuth, invalidate);
			
			// Make some calculations
			lCleanupDone = System.currentTimeMillis();
			long lConnection = lConnectionDone - lExecuteStart;
			long lClient = lClientDone - lConnectionDone;
			long lCommand = lCommandDone - lClientDone;
			long lCleanup = lCleanupDone - lCommandDone;
			
			// Log some events
			if (ts.getCurrentNode() != null) {
				ts.raiseEvent(TestEvent.EVENT_LOGMSG, "SSH Connection Time", lConnection + "");
				ts.raiseEvent(TestEvent.EVENT_LOGMSG, "SFTP Client Creation Time", lClient + "");
				ts.raiseEvent(TestEvent.EVENT_EXECTIME, ts.getCurrentNode().getName(), lCommand + "");
				ts.raiseEvent(TestEvent.EVENT_LOGMSG, "SFTP Client Cleanup Time", lCleanup + "");
			}
		}
	}
	
	/**
	 * 
	 * @param client
	 * @return
	 * @throws IOException
	 */
	@SuppressWarnings("unchecked")
	private String cmd_ls(SFTPv3Client client, String path) throws IOException {
		
		// Create the result value
		StringBuffer sbFiles = new StringBuffer("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		sbFiles.append("<directory path=\"" + path + "\">\n");
		
		final Vector files = client.ls(path);
		for(int i=0; i < files.size(); i++) {
			SFTPv3DirectoryEntry entry = (SFTPv3DirectoryEntry)files.get(i);
			
			// If it contains 0040000 it is a directory
			if ((entry.attributes.permissions.intValue() & 0040000) != 0) {
				sbFiles.append("  <directory>" + entry.filename + "</directory>\n");
			} else {
				sbFiles.append("  <file"
						+ " atime=\"" + convertTime(entry.attributes.atime) + "\""
						+ " atime_long=\"" + entry.attributes.atime + "\""
						+ " mtime=\"" + convertTime(entry.attributes.mtime) + "\""
						+ " mtime_long=\"" + entry.attributes.mtime + "\""
						+ " size=\"" + entry.attributes.size + "\""
						+ ">" + entry.filename + "</file>\n");
			}
		}
		sbFiles.append("</directory>");
		return sbFiles.toString();
	}
	
	/**
	 * 
	 * @param client
	 * @return
	 * @throws IOException
	 */
	private String cmd_read(SFTPv3Client client, String path, String filename, boolean saveToFile, String localFilename) throws IOException {
		
		// Create the result value
		StringBuffer sbFile = new StringBuffer();
		byte[] dst = new byte[32768];
		int fileOffset = 0;
		int len = 0;
		
		// Create the read file handle
		SFTPv3FileHandle handle = client.openFileRO(path + "/" + filename);
		
		// Determine whether to save the file to disk
		FileOutputStream out = null;
		if (saveToFile) {
			out = new FileOutputStream(localFilename);
		}
		
		// Read the file
		while (len != -1) {
			len = client.read(handle, fileOffset, dst, 0, 32768);
			fileOffset += len;
			log.debug("Just read " + len + " bytes over SFTP");
			
			// Get just the bytes we care about
			if ((len < 32768) && (len != -1)) {
				
				// Get only the correct bytes
				byte[] temp = new byte[len];
				System.arraycopy(dst, 0, temp, 0, len);
				
				// Determine if we store to disk or to memory
				if (saveToFile) {
					out.write(temp);
				} else {
					sbFile.append(new String(temp));
				}
			} else {
				
				// Get everything in dst
				// Determine if we store to disk or to memory
				if (saveToFile) {
					out.write(dst);
				} else {
					sbFile.append(new String(dst));
				}
			}
			if (len < 32768) {
				break;
			}
		}
		
		// Cleanup
		client.closeFile(handle);
		if (saveToFile) {
			out.flush();
			out.close();
			sbFile.append("Local file: " + localFilename + " written.");
		}
		return sbFile.toString();
	}
	
	/**
	 * 
	 * @param client
	 * @return
	 * @throws IOException
	 */
	private String cmd_overwrite(SFTPv3Client client, String path, String filename, String content) throws IOException {
		
		// Create the handle
		SFTPv3FileHandle handle = client.openFileRW(path + "/" + filename);
		
		// Write the file
		client.write(handle, 0, content.getBytes(), 0, content.length());
		
		// Cleanup
		client.closeFile(handle);
		
		return path + "/" + filename + " overwritten.";
	}
	
	/**
	 * 
	 * @param client
	 * @return
	 * @throws IOException
	 */
	private String cmd_forcewrite(SFTPv3Client client, String path, String filename, String content) throws IOException {

		String msg;
		try {
			// First try to create/write the file
			msg = cmd_create(client, path, filename, content);
		}
		catch(IOException createIOE) {
			// If the file already exists just overwrite
			msg = cmd_overwrite(client, path, filename, content);
		}
		return msg;
	}
	
	/**
	 * 
	 * @param client
	 * @return
	 * @throws IOException
	 */
	private String cmd_create(SFTPv3Client client, String path, String filename, String content) throws IOException {
		
		// Create the handle
		SFTPv3FileHandle handle = client.createFile(path + "/" + filename);
		
		// Write the file
		client.write(handle, 0, content.getBytes(), 0, content.length());
		
		// Cleanup
		client.closeFile(handle);
		return path + "/" + filename + " created.";
	}
	
	/**
	 * 
	 * @param client
	 * @return
	 * @throws IOException
	 */
	private String cmd_delete(SFTPv3Client client, String path, String filename) throws IOException {
		
		// Delete the file
		try {
			client.rm(path + "/" + filename);
		} catch (IOException ioe) {
			return path + "/" + filename + " was not deleted!";
		}
		return path + "/" + filename + " deleted.";		
	}
	
	private String convertTime(Long l) {
		Date d = new Date(l * 1000);
		return d.toString();
	}
}
