package com.itko.lisa.ext.trilead;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Element;

import com.itko.lisa.ext.trilead.xterm.XtermEngine;
import com.itko.lisa.test.TestCase;
import com.itko.lisa.test.TestDefException;
import com.itko.lisa.test.TestEvent;
import com.itko.lisa.test.TestExec;
import com.itko.lisa.test.TestNode;
import com.itko.lisa.test.TestRunException;
import com.itko.util.CloneImplemented;
import com.itko.util.StrUtil;
import com.itko.util.XMLUtils;
import com.trilead.ssh2.ChannelCondition;
import com.trilead.ssh2.Connection;
import com.trilead.ssh2.Session;
import com.trilead.ssh2.StreamGobbler;

public class SSHStep extends TestNode implements CloneImplemented {

	private static final long serialVersionUID = 7032591440170960620L;
	private static Log log = LogFactory.getLog(SSHStep.class);

	public static final String TAG_HOSTNAME = "hostname";
	public static final String TAG_USERNAME = "username";
	public static final String TAG_PASSWORD = "password";
	public static final String TAG_AUTH = "auth";
	public static final String TAG_PRIVATE_KEY = "privateKey";
	public static final String TAG_PRIVATE_KEY_PASS = "privateKeyPass";
	public static final String TAG_COMMAND = "command";
	public static final String TAG_INT_EXPECT = "intExpect";
	public static final String TAG_INT_COMMAND = "intCommand";
	public static final String TAG_MODE = "mode";
	public static final String TAG_ERROR_NODE = "errorNode";
	
	public static final String NON_INTERACTIVE_MODE = "Non-Interactive Command";
	public static final String INTERACTIVE_MODE = "Interactive Login Shell";
	public static final String[] MODE_ARR = {
		NON_INTERACTIVE_MODE,
		INTERACTIVE_MODE
	};

	private String hostname;
	private String username;
	private String password;
	private String auth;
	private String privateKey;
	private String privateKeyPass;

	private String command;
	private String intCommand;
	private String intExpect;
	private String mode;
	private String errorNode;

	public String getHostname() {
		return hostname;
	}

	public void setHostname(String hostname) {
		this.hostname = hostname;
	}

	public String getUsername() {
		return username;
	}

	public void setUsername(String username) {
		this.username = username;
	}

	public String getPassword() {
		return password;
	}

	public void setPassword(String password) {
		this.password = password;
	}

	public String getAuth() {
		return auth;
	}

	public void setAuth(String auth) {
		if (auth == null) {
			auth = TrileadConnectionFactory.AUTH_PASSWORD;
		} else {
			this.auth = auth;
		}
	}

	public String getPrivateKey() {
		return privateKey;
	}

	public void setPrivateKey(String privateKey) {
		this.privateKey = privateKey;
	}

	public String getPrivateKeyPass() {
		return privateKeyPass;
	}

	public void setPrivateKeyPass(String privateKeyPass) {
		this.privateKeyPass = privateKeyPass;
	}
	
	public String getCommand() {
		return command;
	}

	public void setCommand(String command) {
		this.command = command;
	}
	
	public String getIntCommand() {
		return intCommand;
	}
	
	public void setIntCommand(String intCommand) {
		this.intCommand = intCommand;
	}
	
	public String getIntExpect() {
		return intExpect;
	}
	
	public void setIntExpect(String intExpect) {
		this.intExpect = intExpect;
	}
	
	public String getMode() {
		return mode;
	}
	
	public void setMode(String mode) {
		if (mode == null) {
			mode = NON_INTERACTIVE_MODE;
		}
		this.mode = mode;
	}
	
	public String getErrorNode() {
		if (errorNode == null) {
			return "fail";
		}
		return errorNode;
	}
	
	public void setErrorNode(String errorNode) {
		if (errorNode == null) {
			this.errorNode = "fail";
		}
		this.errorNode = errorNode;
	}


	@Override
	protected void execute(TestExec ts) throws TestRunException {
		
		// Timers
		long lExecuteStart = System.currentTimeMillis();
		long lConnectionDone = System.currentTimeMillis();
		long lClientDone = System.currentTimeMillis();
		long lCommandDone = System.currentTimeMillis();
		long lCleanupDone = System.currentTimeMillis();

		// Parsed Properties
		String tsHostname = ts.parseInState(hostname);
		String tsUsername = ts.parseInState(username);
		String tsPassword = ts.parseInState(password);
		String tsCommand = ts.parseInState(command);
		String tsAuth = ts.parseInState(auth);
		String tsPrivateKey = ts.parseInState(privateKey);
		String tsPrivateKeyPass = ts.parseInState(privateKeyPass);
		
		// Connections
		Connection conn = null;
		Session sess = null;
		boolean invalidate = true;
		
		try {
			// Get the connection
			conn = TrileadConnectionFactory.getConnection(tsHostname, tsUsername, tsPassword, tsAuth, tsPrivateKey, tsPrivateKeyPass);
			lConnectionDone = System.currentTimeMillis();
			
			// Create a session
			sess = conn.openSession();
			lClientDone = System.currentTimeMillis();
			
			// Determine whether to run the command or terminal session
			String msg = new String("");
			if (mode.equals(NON_INTERACTIVE_MODE)) {
				
				// Run the command
				SSHResult result = execSSHCommand(sess, tsCommand);
				lCommandDone = System.currentTimeMillis();
				
				// Check the return code
				msg = result.getStdout();
				if (result.getExitStatus() != null) {
					if (result.getExitStatus().intValue() != 0) {
						msg = result.getStderr();
					}
				} else {
					// Make sure we show an error
					result.setExitStatus(-1);
				}
				
				// Set properties
				String prop = "lisa." + ts.getCurrentNodeName() + ".";
				ts.setStateValue(prop + "stdout", result.getStdout());
				ts.setStateValue(prop + "stderr", result.getStderr());
				ts.setStateValue(prop + "exitstatus", result.getExitStatus());
				ts.setStateValue("ssh.last.exitstatus", result.getExitStatus());
				
			} else {
				
				// Get the engine
				XtermEngine engine = XtermEngine.getEngine(tsHostname, tsUsername, tsAuth);
				engine.connect(sess);
				engine.setEngineStatus(XtermEngine.STATUS_PLAYBACK);
				
			}
			
			ts.setLastResponse(msg);
			ts.raiseEvent(TestEvent.EVENT_NODERESPONSE, ts.getCurrentNode().getName(), msg);

			invalidate = false;
		} catch (Exception e) {
            ts.setLastResponse(e.getMessage());
            raiseNodeFailedEvent(ts, getErrorNode(), name, e.getMessage(), e);
		} finally {
			
			// Close the session 
			if (sess != null) {
				sess.close();
			}
			
			// Toss the connection
			TrileadConnectionFactory.cleanupConnection(conn, tsHostname, tsUsername, tsAuth, invalidate);
			
			// Make some calculations
			lCleanupDone = System.currentTimeMillis();
			long lConnection = lConnectionDone - lExecuteStart;
			long lClient = lClientDone - lConnectionDone;
			long lCommand = lCommandDone - lClientDone;
			long lCleanup = lCleanupDone - lCommandDone;
			
			// Log some events
			if (ts.getCurrentNode() != null) {
				ts.raiseEvent(TestEvent.EVENT_LOGMSG, "SSH Connection Time", lConnection + "");
				ts.raiseEvent(TestEvent.EVENT_LOGMSG, "SFTP Client Creation Time", lClient + "");
				ts.raiseEvent(TestEvent.EVENT_EXECTIME, ts.getCurrentNode().getName(), lCommand + "");
				ts.raiseEvent(TestEvent.EVENT_LOGMSG, "SFTP Client Cleanup Time", lCleanup + "");
			}

		}
	}
	
//	public static String execSSHCommand(Session sess, String tsCommand) throws Exception{
	public static SSHResult execSSHCommand(Session sess, String tsCommand) throws Exception {
	
		SSHResult result = new SSHResult();
		
		// Create a session and run the command
		sess.execCommand(tsCommand);
		
		// Wait up to 30 seconds until we hit the EOF
		sess.waitForCondition(ChannelCondition.EOF, 30000);
		
		// Read the stream
		InputStream stdout = new StreamGobbler(sess.getStdout());
		BufferedReader brStdout = new BufferedReader(new InputStreamReader(stdout));
		
		// Read the error stream
		InputStream stderr = new StreamGobbler(sess.getStderr());
		BufferedReader brStderr = new BufferedReader(new InputStreamReader(stderr));
		
		// Read the stdout
		result.setStdout(readBuffer(brStdout));
		log.debug("stdout = " + result.getStdout());
		
		// Read the stderr
		result.setStderr(readBuffer(brStderr));
		log.debug("stderr = " + result.getStderr());
		
		// Grab the return code
		result.setExitStatus(sess.getExitStatus());
		log.debug("exit status = " + result.getExitStatus());
		if (result.getExitStatus() == null) {
			
			// Put a status of -1 (UNKNOWN)
			result.setExitStatus(-1);
		}
		
		return result;
	}

	/**
	 * Helper utility to read all the data out of the buffer (stdout or stderr)
	 * 
	 * @param br
	 * @return
	 * @throws IOException
	 */
	private static String readBuffer(BufferedReader br) throws IOException {
		
		StringBuffer sb = new StringBuffer();
		while (true) {
			String line = br.readLine();
			if (line == null) {
				break;
			}
			sb.append(line + "\n");
		}
		return sb.toString();
	}
	
	@Override
	public void initialize(TestCase test, Element node) throws TestDefException {

		setHostname(XMLUtils.findChildGetItsText(node, TAG_HOSTNAME));
		setUsername(XMLUtils.findChildGetItsText(node, TAG_USERNAME));
		setAuth(XMLUtils.findChildGetItsText(node, TAG_AUTH));
		setPrivateKey(XMLUtils.findChildGetItsText(node, TAG_PRIVATE_KEY));
		
		setCommand(XMLUtils.findChildGetItsText(node, TAG_COMMAND));
		setIntCommand(XMLUtils.findChildGetItsText(node, TAG_INT_COMMAND));
		setIntExpect(XMLUtils.findChildGetItsText(node, TAG_INT_EXPECT));
		setMode(XMLUtils.findChildGetItsText(node, TAG_MODE));
		setErrorNode(XMLUtils.findChildGetItsText(node, TAG_ERROR_NODE));
		
		// Decode the regular password
		try {
			String pass = XMLUtils.findChildGetItsText(node, TAG_PASSWORD);
			setPassword(StrUtil.decodeString(pass));
		} catch (IOException ioe) {
			setPassword("");
		}
		
		// Decode the private key password
		try {
			String pass2 = XMLUtils.findChildGetItsText(node, TAG_PRIVATE_KEY_PASS);
			setPrivateKeyPass(StrUtil.decodeString(pass2));
		} catch (IOException ioe) {
			setPrivateKeyPass("");
		}
	}
	
	public void writeSubXML( PrintWriter pw )
    {
    	XMLUtils.streamTagAndChild(pw, TAG_HOSTNAME, hostname);
    	XMLUtils.streamTagAndChild(pw, TAG_USERNAME, username);
    	XMLUtils.streamTagAndChild(pw, TAG_AUTH, auth);
    	XMLUtils.streamTagAndChild(pw, TAG_PRIVATE_KEY, privateKey);
    	XMLUtils.streamTagAndChild(pw, TAG_COMMAND, command);
    	XMLUtils.streamTagAndChild(pw, TAG_INT_COMMAND, intCommand);
    	XMLUtils.streamTagAndChild(pw, TAG_INT_EXPECT, intExpect);
    	
    	XMLUtils.streamTagAndChild(pw, TAG_MODE, mode);
    	XMLUtils.streamTagAndChild(pw, TAG_ERROR_NODE, errorNode);
    	
    	String tmp = StrUtil.encodeString(password);
    	XMLUtils.streamTagAndChild(pw, TAG_PASSWORD, tmp);
    	
    	String tmp2 = StrUtil.encodeString(privateKeyPass);
    	XMLUtils.streamTagAndChild(pw, TAG_PRIVATE_KEY_PASS, tmp2);
    }
	   
	public String getTypeName() throws Exception {
		return "SSH Step";
	}

}
